#!/usr/bin/wish

proc updateInsertTag { Text xml enTagName lAllowNode paramInsertTag } {
  tset cur_node [richtext::Cursor $Text get TAG]
  switch -- $paramInsertTag {
    before -
    after  {
      if {[simxml::nodeInfo $xml [tset cur_node] type] == "TEXT" } {
        tset cur_node [simxml::nodeInfo $xml [tset cur_node] parent]
      } ;# if
      if { [tset cur_node] != [simxml::root get $xml] } {
        tset cur_node [simxml::nodeInfo $xml [tset cur_node] parent]
      } ;# if
      tset cur_node [simxml::nodeInfo $xml [tset cur_node] name]
      tset listTag [simxml::dtdListTag $xml [tset cur_node]]
    }
    inside {
      if {[simxml::nodeInfo $xml [tset cur_node] type] == "TEXT" } {
        tset cur_node [simxml::nodeInfo $xml [tset cur_node] parent]
      } ;# if
      tset cur_node [simxml::nodeInfo $xml [tset cur_node] name]
      tset listTag [simxml::dtdListTag $xml [tset cur_node]]
    }
  } ;# switch ( before | inside | after )

  if { [lsearch [tset listTag] _ALL_] != -1 } {
    $enTagName configure -state normal
    tset listTag [lreplace [tset listTag] [lsearch [tset listTag] _ALL_] [lsearch [tset listTag] _ALL_]]
  } else {
    $enTagName configure -state normal
  } ;# if-else
  $lAllowNode delete 0 end
  eval $lAllowNode insert 0 [tset listTag]
  tunset listTag
} ;# updateInsertTag




#============================================================================
#============================================================================
#============================================================================
proc gui::prNewFile {} {
         set Text [gui::data get currentTextWidget]

         set info_doc [gui::Dialog [gui::data get guiXML] newXMLDocument]

         if { [string equal $info_doc {}] } { return }
         catch { gui::prCloseFile }
         set xml  [simxml::create -encoding [lindex $info_doc 1]]
         simxml::add $xml {} [lindex $info_doc 0] -type TAG
         set html [simxml::root get $xml ]
         set css  [simcss::create ]

         richtext::widgetData $Text set auto-indent 10
         richtext::widgetData $Text set css         $css
         richtext::widgetData $Text set xml         $xml
         richtext::widgetData $Text set xml_root    $html
         richtext::widgetData $Text set textModify  1
         focus $Text
         wm title . "New XML document | [gui::data get tkXMLiVEversion]"
         .menubar.file entryconfigure [gui::menuGet .menubar.file close]  -state normal
         .menubar.file entryconfigure [gui::menuGet .menubar.file save]   -state disabled
         .menubar.file entryconfigure [gui::menuGet .menubar.file saveAs] -state normal
         .menubar.file entryconfigure [gui::menuGet .menubar.file xslTransform] -state normal
         .menubar.file entryconfigure [gui::menuGet .menubar.file saveDTD]   -state normal

         set fileCSSName [lindex $info_doc 2]
         if { ![string equal $fileCSSName {}] } {
           simxml::PrCommand $xml set XML-STYLESHEET [list type text/css href $fileCSSName]
#           set css  [simcss::create -file $fileCSSName]
           gui::prOpenStyle $fileCSSName
         } ;# if

  gui::TreeExpressFill
  richtext::updateWidget $Text {}
} ;# proc gui::prNewFile

proc gui::prOpenFile {} {
      set Text     [gui::data get currentTextWidget]

      if { [richtext::widgetData $Text get textModify] } {
        switch -- [gui::Dialog {} dlg_WarmCloseFile] {
          1  {  gui::prSaveFile }
          -1 {  return          }
        } ;# switch
      } ;# if


         set fileName [tk_getOpenFile -title [mc "Open XML file"]  -filetypes { \
                      {{XML documents} {.xml .XML}}
                      {{XSL documents} {.xsl .XSL .xslt .XSLT}}
                      {{All documents} {*}}}];

         if { ![string equal $fileName {}] } {
           richtext::widgetData $Text set textModify 0
           set isClose 0
           catch { gui::prCloseFile    } isClose
           if { $isClose == -1 } { return }
           catch { simxml::delete $xml }

           set xml  [simxml::parseXML -file $fileName]
           set html [simxml::root get $xml ]
           set css  [simcss::create]
           richtext::widgetData $Text set auto-indent 10
           richtext::widgetData $Text set textModify  0

# BUG !!!!!!!!!!!! need if check PrCommand
           set fileCSSName {}
           catch {
             foreach { tmp } [simxml::PrCommand $xml get XML-STYLESHEET] {
               array set tmp_style $tmp
               set fileCSSName [file join [file dirname $fileName] $tmp_style(href)]
             } ;# foreach
           } ;# catch

           richtext::widgetData   $Text set css      $css
           richtext::widgetData   $Text set xml      $xml
           richtext::widgetData   $Text set xml_root $html
           focus $Text
           wm title . "[simxml::xmlInfo $xml get file] | [gui::data get tkXMLiVEversion]"
           .menubar.file entryconfigure [gui::menuGet .menubar.file close]  -state normal
           .menubar.file entryconfigure [gui::menuGet .menubar.file save]   -state normal
           .menubar.file entryconfigure [gui::menuGet .menubar.file saveAs] -state normal
           .menubar.file entryconfigure [gui::menuGet .menubar.file xslTransform] -state normal
           .menubar.file entryconfigure [gui::menuGet .menubar.file saveDTD]   -state normal
           gui::msgDebug "USED style: $fileCSSName"
           if { ! [string equal $fileCSSName {}] } {
             gui::prOpenStyle $fileCSSName
           } else {
             richtext::updateWidget $Text {}
           } ;# if-else
        }
      gui::TreeExpressFill
} ;# proc gui::prOpenFile 


proc gui::prSaveFile { args } {
      set Text     [gui::data get currentTextWidget]
      set xml      [richtext::widgetData $Text get xml]
      set fileName [simxml::xmlInfo $xml get file]

      if { ! [string equal [lindex $args 0] {}] } {
        set fileName [lindex $args 0]
      } ;# if

      if { [string equal $fileName {}] } {
        gui::prSaveAsFile
        return
      } ;# if

      if { [catch {open $fileName w} OUT] } {
        gui::msgDebug "Error write file $fileName"
      } else {
        set Text [gui::data get currentTextWidget]
        set xml  [richtext::widgetData $Text get xml]
        fconfigure $OUT -encoding [simxml::xmlInfo $xml get encoding]
          puts $OUT [simxml::debug $xml {} -recurse true]
        close $OUT
        richtext::widgetData $Text set textModify 0
      } ;# if
} ;# proc gui::prSaveFile


proc gui::prSaveAsFile {} {
      set fileName [tk_getSaveFile -title [mc "Save as... XML file"]  -filetypes { \
                                           {{XML documents}  {.xml .XML}}
                                           {{XSLT documents} {.xsl .XSL .xslt .XSLT}}
                                           {{All documents}  {*}}}]
      gui::prSaveFile $fileName
} ;# proc gui::prSaveAsFile


proc gui::prCloseFile {} {
      set Text [gui::data get currentTextWidget]
      set xml  [richtext::widgetData $Text get xml]

      if { [richtext::widgetData $Text get textModify] } {
        switch -- [gui::Dialog {} dlg_WarmCloseFile] {
          1  {  gui::prSaveFile }
          -1 {  return -1       }
        } ;# switch
      } ;# if

      eval {$Text mark unset} [$Text mark names]
      simxml::delete $xml
      $Text delete 1.0 end
      richtext::widgetData $Text set xml         {}
      richtext::widgetData $Text set css         {}
      richtext::widgetData $Text set textModify  0
      set xml {}
      wm title . [gui::data get tkXMLiVEversion]
      .menubar.file entryconfigure [gui::menuGet .menubar.file close]  -state disabled
      .menubar.file entryconfigure [gui::menuGet .menubar.file save]   -state disabled
      .menubar.file entryconfigure [gui::menuGet .menubar.file saveAs] -state disabled
      .menubar.file entryconfigure [gui::menuGet .menubar.file xslTransform] -state disabled
      .menubar.file entryconfigure [gui::menuGet .menubar.file saveDTD]   -state disabled

      $::Tree delete [$::Tree nodes root]
      return 1
} ;# proc gui::prCloseFile


proc gui::prOpenStyle { args } {
        set fileName [lindex $args 0]

        if { [string equal $fileName {}] } {
          set fileName [tk_getOpenFile -title "Open XML file" \
                                       -filetypes {{{CSS files}  {.css}} \
                                                  {{All files} {*}} \
                                     }]
        } ;# if

        if { ![string equal $fileName {}] } {
          set styleName [file tail [file rootname $fileName]]
          set menu_attr [list name     style$::countStyle \
                              label    $styleName \
                              accel    Control-$::countStyle \
                              command  "gui::prSetStyle $fileName"]
          set newXML [simxml::create]
            set new_node [simxml::add $newXML {} COMMAND -type element -attribute $menu_attr]
            gui::menuAdd .menubar.style $newXML $new_node
          simxml::delete $newXML

          incr ::countStyle
          set ::varShowTAG OFF
          gui::prSetStyle $fileName
        } ;# if fileName != ""
} ;# proc gui::prOpenStyle


proc gui::prSetStyle { fileName } {
    set Text [gui::data get currentTextWidget]
    if { ![string equal $fileName {}] } {
      set css  [simcss::create -file $fileName]
      richtext::widgetData $Text set auto-indent 0
      richtext::widgetData $Text set show-tag    OFF
      richtext::widgetData $Text set css         $css
      gui::msgDebug "UPDATE TIME: [time {richtext::updateWidget $Text {}}]"
    } ;# if
} ;# proc gui::prSetStyle


proc gui::prDefaultStyle {} {
    set Text [gui::data get currentTextWidget]
    set css  [simcss::create]
    richtext::widgetData $Text set auto-indent 10
    richtext::widgetData $Text set show-tag    AUTO
    richtext::widgetData $Text set css         $css
    richtext::updateWidget $Text {}
} ;# proc gui::prDefaultStyle


proc gui::prChangeXMLWidget { param value } {
    set Text [gui::data get currentTextWidget]
    richtext::widgetData $Text set $param $value
    set cur_pos [$Text index [richtext::Cursor $Text get]]
    richtext::updateWidget $Text {}
    richtext::Cursor $Text set $cur_pos
    $Text see insert
} ;# proc gui::prChangeXMLWidget


proc gui::prXSLTransform {} {
  gui::Dialog {} dlg_XSLTransform
} ;# proc gui::prXSLTransform


proc gui::TreeFill { widget xml node_id font } {
  set node_parent [simxml::nodeInfo $xml $node_id parent]
  set node_name   [simxml::nodeInfo $xml $node_id name]
  set node_type   [simxml::nodeInfo $xml $node_id type]
  set text        $node_name

  set node_children [simxml::children $xml $node_id -verbose expanded]
  if { [set child_index [lsearch -glob $node_children *TITLE]] != -1 } {
    set children [lindex $node_children [expr $child_index - 1]]
    set text "$text ([simxml::getText $xml $children])"
  } ;# if

  switch -- $node_type {
    ELEMENT -
    TAG     {
      if { [string equal $node_parent {}] } { set node_parent root }
      $widget insert end $node_parent $node_id -text $text -open 0 -font $font
    }
  } ;# switch
} ;# proc gui::TreeFill


proc gui::TreeNodeClick { widget node } {
  richtext::Cursor [gui::data get currentTextWidget] set TAG $node
  [gui::data get currentTextWidget] see insert
  $widget selection set $node
  focus [gui::data get currentTextWidget]
} ;# proc gui::TreeNodeClick


proc gui::TreeNodeDClick { widget node } {
  richtext::Cursor [gui::data get currentTextWidget] set TAG $node
  [gui::data get currentTextWidget] see insert
  if { [$widget itemcget $node -open] == 0 } {
    $widget itemconfigure $node -open 1
  } else {
    $widget itemconfigure $node -open 0
  } ;# if
  $widget selection set $node
  focus [gui::data get currentTextWidget]
} ;# proc gui::TreeNodeDClick


proc gui::TreeNodeEnter { widget w node } {
  #set text [$widget itemcget $node -text]
  #gui::balloon $w show "$text"
} ;# proc gui::TreeNodeEnter


proc gui::TreeNodeLeave { widget w node } {
} ;# proc gui::TreeNodeLeave


proc gui::TreeExpressFill { args } {
  set Text [gui::data get currentTextWidget]
  set xml  [richtext::widgetData $Text get xml]
  $::Tree delete [$::Tree nodes root]
  simxml::recurseTree $xml -root {} \
                      -command [list gui::TreeFill $::Tree %XML %NODE_ID {Arial 9}]

  if  { [string equal [lindex $args 0] {}] } {
    set xml_root [richtext::widgetData $Text get xml_root]
  } else {
    set xml_root [lindex $args 0]
  }

  foreach { node } [simxml::listNodePath $xml $xml_root] {
    $::Tree itemconfigure $node -open 1
  } ;# foreach

  $::Tree selection set $xml_root
  $::Tree see $xml_root
} ;# proc gui::TreeExpressFill 


proc gui::TreeUpdate { xml node action } {
  switch -- [simxml::nodeInfo $xml $node type] {
    TEXT {  set node [simxml::nodeInfo $xml $node parent]  }
  } ;# switch

  foreach { tnode } [simxml::listNodePath $xml $node] {
    catch {
      if { [$::Tree itemcget $tnode -open] == 0 } {  break  } ;# if
    }
  } ;# foreach
  $::Tree selection set $tnode
  $::Tree see $tnode
} ;# proc gui::TreeUpdate


proc gui::prChangeShowDebug { } {
  if { $::varShowDebug == "ON" } {
    pack [ gui::data get debugWidget ] -side top  -fill x
  } else {
    pack forget [ gui::data get debugWidget ]
  };# if
} ;# proc gui::prChangeShowDebug




#=================================
#=================================
#=================================


$splash.status configure -text [mc "Loading GUI XML..."]
update idletasks
set guiXML [simxml::parseXML -file gui/gui.xml]


set ButtonBox [frame .bbox -height 32  -borderwidth 1 -relief groove -background #EEEEEE]

$splash.status configure -text [mc "Create ToolBar..."]
update idletasks
gui::toolbarCreate $ButtonBox -xml $guiXML
pack $ButtonBox -side top -expand false -fill x


$splash.status configure -text [mc "Create User Interface..."]
update idletasks
set wRoot [gui::createUI . -xml $guiXML]
pack [lindex $wRoot 0] -side top -fill both



gui::data set currentTextWidget $Text
set paramInsertTag inside
set countStyle     1
set varShowTAG     AUTO
set varShowDebug   ON
set varFontSize    Normal
set varIndentSize  Normal

gui::prChangeShowDebug


gui::menuCreate . $guiXML menubar
SetResource .


;# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
;#                         USER DEFINED EVENTS
;# =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
if { 1 == 1} {
richtext::Bind $Text after <<CursorMove>> {
  richtext::wRTStatus:extStatus [richtext::Cursor %W get TAG]
} ;# BIND after <<CursorMove>

richtext::Bind $Text after <<CursorMove>> {
# HightLight position
    %W tag remove hightlightBorder 1.0 end
    tset node_cur    [richtext::Cursor %W get TAG]
    switch -- [simxml::nodeInfo [richtext::widgetData %W get xml] \
                                [tset node_cur] type] {
      ELEMENT {
        tset node_index [richtext::tagIndex %W [tset node_cur] -type element]
        %W tag add hightlightBorder [lindex [tset node_index] 0] \
                                    [lindex [tset node_index] 1]
        %W tag raise hightlightBorder
      }
      TEXT {
        tset node_index [richtext::tagIndex %W [tset node_cur] -type text]
        %W tag add hightlightBorder [lindex [tset node_index] 0] \
                                    [lindex [tset node_index] 1]
        %W tag raise hightlightBorder
      }
      TAG {
        tset node_start [richtext::tagIndex %W [tset node_cur] -type open]
        tset node_end   [richtext::tagIndex %W [tset node_cur] -type close]
        %W tag add hightlightBorder [lindex [tset node_start] 0] \
                                    [lindex [tset node_end  ] 1]
        %W tag raise hightlightBorder
      }
    } ;# switch
    updateInsertTag %W [richtext::widgetData %W get xml] \
                    $::enTagName  $::lAllowNode  $::paramInsertTag

  gui::TreeUpdate [richtext::widgetData %W get xml] [tset node_cur] {}
} ;# BIND after <<CursorMove>
} ;# if COMMENT








