package provide gui 1.0

namespace eval gui {
  variable xml
  variable ID 0
  variable BALLOON
  variable WIDGET    ;  array set WIDGET    {}
  variable DATA      ;  array set DATA      {}
  variable dialogVar ;  array set dialogVar {}
  variable xmlNODE   ;  array set xmlNODE   {}
  variable menuINDEX ;  array set menuINDEX {}
  variable toolBar   ;  array set toolBar   {}

  proc initialize { args } {
  variable xmlNODE
    set a_args(-file) {}
    set a_args(-xml)  {}
    array set a_args $args

    if { [string equal $a_args(-xml) {}] } \
    { set xml [simxml::parseXML -file $a_args(-file)] } else { set xml $a_args(-xml) }

    set xmlNODE(init)    true
    set xmlNODE(STYLES)   [simxml::getNodes $xml {} name STYLES]
    set xmlNODE(ROOT)    [simxml::getNodes $xml {} name ROOT]
    set xmlNODE(DIALOGS) [simxml::getNodes $xml {} name DIALOGS]

# CREATE procedure
    set root_node [simxml::getNodes $xml {} name PROCEDURES -one-level-child true]
    foreach { prNode } [simxml::nodeInfo $xml $root_node children] {
      switch -- [simxml::nodeInfo $xml $prNode name] {
        PROCEDURE {
          set prAttr   {}
          set prName   [simxml::attribute $xml get $prNode -name id]
  catch { set prAttr   [simxml::attribute $xml get $prNode -name param] }
          set prSource [simxml::getText $xml $prNode]
          eval "proc [namespace current]::$prName { $prAttr } { $prSource } "
        }
      } ;# switch name node
    } ;# foreach all procedures

  } ;# proc initialize

  proc createUI { parent args } {
  variable xml
  variable WIDGET
  variable xmlNODE
    set a_args(-file) {}
    set a_args(-xml)  {}
    set a_args(-root) ROOT
    set a_args(-node) {}
    array set a_args $args

    if { $parent == "." }  { set parent {} }

    if { [string equal $a_args(-xml) {}] } \
    { set xml [simxml::parseXML -file $a_args(-file)] } else { set xml $a_args(-xml) }

    gui::data set guiXML $xml
    initialize -xml $xml

    set root_node [simxml::getNodes $xml {} name $a_args(-root) -one-level-child true]
    if { ! [string equal $a_args(-node) {}] } {  set root_node $a_args(-node)  }

    foreach { rNode } $root_node {
      array set rAttr [simxml::attribute $xml get $rNode -param true]
      set rWidget $rAttr(-id)  ;  unset rAttr(-id)
      set wParent [eval frame $parent.$rWidget [array get rAttr]]

      foreach { Widget } [simxml::nodeInfo $xml $rNode children] {
        set paneWidget [createWidget $wParent $Widget]
        eval pack [lindex $paneWidget 0] -expand true -fill both [lindex $paneWidget 1]
      } ;# foreach

      return $wParent
    } ;# foreach
  } ;# proc create


  proc createWidget { wParent node args } {
    variable xml
    variable WIDGET
    set a_args(-widget)   {}
    set a_args(-type)     {}
    array set a_args $args

    if { [string equal $a_args(-widget) {}] } {
      set a_args(-widget) [simxml::attribute $xml get $node -name id]
    } ;# if
    if { [string equal $a_args(-type)   {}] } {
      set a_args(-type)   [simxml::nodeInfo $xml $node name]
    } ;# if
    set wText [simxml::getText $xml $node]

    array set rAttr {}
    array set rAttr [simxml::attribute $xml get $node -param true]
    catch { unset rAttr(-id) }

    if [info exists rAttr(-style)] {
      array set tmpAttr [getStyle $xml $rAttr(-style)]
      array set tmpAttr [array get rAttr]
      array set rAttr   [array get tmpAttr]
      unset rAttr(-style)
    } ;# if STYLE present

    set wVariable {} ; catch { set wVariable $rAttr(-variable) }
    catch { unset rAttr(-variable) }
    set wPack {} ; catch { set wPack $rAttr(-pack) }
    catch { unset rAttr(-pack) }

    switch -- $a_args(-type) {
      PANE {
        set frLeft [ set frRight {} ]
        foreach { frLeft frRight } [simxml::nodeInfo $xml $node children] { break }
        set wLeft  [lindex [createWidget $wParent $frLeft ] 0]
        set wRight [lindex [createWidget $wParent $frRight] 0]
        pack $wLeft  -expand true -fill both
        pack $wRight -expand true -fill both
        eval pane::create $wLeft $wRight -in $wParent [array get rAttr]
        return [list $wParent $wPack]
      }
      FRAME {
        set frWidget [eval frame $wParent.$a_args(-widget) [array get rAttr]]
        foreach { frChild } [simxml::nodeInfo $xml $node children] {
          set fr_info [createWidget $frWidget $frChild]
          eval pack [lindex $fr_info 0] -expand true -fill both [lindex $fr_info 1]
        }
        return [list $frWidget $wPack]
      }
      LABEL {
        if [info exists rAttr(-image)] {
          set rAttr(-image) [image create photo -file $rAttr(-image)]
        } ;# if IMAGE attribute
        set frWidget [eval {label $wParent.$a_args(-widget) -text $wText } [array get rAttr]]

        foreach { listItem } [simxml::getNodes $xml $node name BIND] {
          bind $frWidget <[simxml::attribute $xml get $listItem -name event]> [simxml::getText $xml $listItem]
        } ;# foreach BIND
        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frWidget $wPack]
      }
      TEXT {
        set frWidget [eval {text $wParent.$a_args(-widget) } [array get rAttr]]
        $frWidget insert 1.0 $wText

        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frWidget $wPack]
      }
      RICHTEXT {
        set frTextBox [frame $wParent.richtext]
        set stToolBar {}
        if [info exists rAttr(-scrollbar)] {
          if [regexp {^style\((.+?)\)$} $rAttr(-scrollbar) -> style_id] {
            set styles_node [simxml::getNodes $xml {} name STYLES]
            foreach { st_child } [simxml::getNodes $xml $styles_node name ITEM] {
            if {[string equal [simxml::attribute $xml get $st_child -name id] $style_id] } {
              array set stToolBar1 [simxml::attribute $xml get $st_child -param true]
              unset stToolBar1(-id)
              set stToolBar [array get stToolBar1]
              unset stToolBar1
              break
            } ;# if
            } ;# foreach
          } else {
            set stToolBar $rAttr(-scrollbar);
          } ;# if-else
        } ;# if
        unset rAttr(-scrollbar)
        set frText    [eval {richtext::create $frTextBox -widget $a_args(-widget) \
                                        -yscrollcommand "$frTextBox.yscroll set"} \
                                        [array get rAttr] ]

#        -height 10 -setgrid true -relief flat \
#                                 -background white

        eval {scrollbar $frTextBox.yscroll -orient vertical -command "$frText yview"} \
                                          $stToolBar

        grid $frText    $frTextBox.yscroll   -sticky news
        grid rowconfigure    $frTextBox 0   -weight 1
        grid columnconfigure $frTextBox 0   -weight 1

        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frText   }
        return [list $frTextBox "-expand true -fill both"]
      }
      STATUS {
        set StatusBox [frame $wParent.statusbox]
        set frStatus  [eval {label $StatusBox.$a_args(-widget) -text $wText } [array get rAttr]]
        grid $frStatus  -sticky wsn
        grid rowconfigure    $StatusBox 0 -weight 1
        grid columnconfigure $StatusBox 0 -weight 1

        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frStatus  }
        return [list $StatusBox $wPack]
      }
      BUTTON {
        set frCommand {}
        catch {
          set frCommand [simxml::getText $xml [simxml::getNodes $xml $node name COMMAND]]
        } ;# catch
        set frWidget [eval {button $wParent.$a_args(-widget) -text $wText -command $frCommand} \
                                   [array get rAttr]]

        foreach { listItem } [simxml::getNodes $xml $node name BIND] {
          bind $frWidget <[simxml::attribute $xml get $listItem -name event]> \
                          [simxml::getText $xml $listItem]
        } ;# foreach BIND

        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frWidget $wPack]
      }
      LISTBOX {
        set frWidget [eval {listbox $wParent.$a_args(-widget)} [array get rAttr]]

        foreach { listItem } [simxml::getNodes $xml $node name BIND] {
          bind $frList <[simxml::attribute $xml get $listItem -name event]> [simxml::getText $xml $listItem]
        } ;# foreach BIND
        foreach { listItem } [simxml::getNodes $xml $node name ITEM] {
          $frWidget insert end [simxml::getText $xml $listItem]
        } ;# foreach

        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frWidget $wPack]
      }
      SCROLLLISTBOX {
        set frListBox [frame $wParent.slb:$a_args(-widget)]
        set stToolBar {}
        if [info exists rAttr(-toolbar)] {
          if [regexp {^style\((.+?)\)$} $rAttr(-toolbar) -> style_id] {
            set styles_node [simxml::getNodes $xml {} name STYLES]
            foreach { st_child } [simxml::getNodes $xml $styles_node name ITEM] {
            if {[string equal [simxml::attribute $xml get $st_child -name id] $style_id] } {
              array set stToolBar1 [simxml::attribute $xml get $st_child -param true]
              unset stToolBar1(-id)
              set stToolBar [array get stToolBar1]
              unset stToolBar1
              break
            } ;# if
            } ;# foreach
          } else {
            set stToolBar $rAttr(-toolbar);
          } ;# if-else
        } ;# if
        unset rAttr(-toolbar)
        set frWidget  [eval {listbox $frListBox.$a_args(-widget) \
                                     -yscrollcommand "$frListBox.yscroll set" \
                                     -xscrollcommand "$frListBox.xscroll set"} \
                                     [array get rAttr] ]
        eval {scrollbar $frListBox.yscroll -orient vertical -command "$frWidget yview"} \
                                           $stToolBar
        eval {scrollbar $frListBox.xscroll -orient horizontal -command "$frWidget xview"} \
                                           $stToolBar

        grid $frWidget    $frListBox.yscroll   -sticky news
        grid $frListBox.xscroll                -sticky new
        grid rowconfigure    $frListBox 0      -weight 1
        grid columnconfigure $frListBox 0      -weight 1
        foreach { listItem } [simxml::getNodes $xml $node name BIND] {
          bind $frWidget <[simxml::attribute $xml get $listItem -name event]> [simxml::getText $xml $listItem]
        } ;# foreach BIND
        foreach { listItem } [simxml::getNodes $xml $node name ITEM] {
          $frWidget insert end [simxml::getText $xml $listItem]
          catch {
          if [string equal [simxml::attribute $xml get $listItem -name select] true] {
            $frWidget selection set end
          } ;# if select is true
          } ;# catch
        } ;# foreach ITEM

        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frListBox $wPack]
      }
      SCROLLBAR {
        set frWidget [eval {scrollbar $wParent.$a_args(-widget)} [array get rAttr]]
        foreach { listItem } [simxml::getNodes $xml $node name BIND] {
          bind $frWidget <[simxml::attribute $xml get $listItem -name event]> [simxml::getText $xml $listItem]
        } ;# foreach BIND
        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frWidget $wPack]
      }
      ENTRY {
        set frWidget [eval {entry $wParent.$a_args(-widget)} [array get rAttr]]
        foreach { listItem } [simxml::getNodes $xml $node name BIND] {
          bind $frWidget <[simxml::attribute $xml get $listItem -name event]> [simxml::getText $xml $listItem]
        } ;# foreach BIND
        if { ! [string equal $wVariable {}] } {  set ::$wVariable $frWidget  }
        return [list $frWidget $wPack]
      }
      CREATECODE {
        set source_code [simxml::getText $xml $node]
        regsub -all {([^%])%F} $source_code "\\1$wParent"            source_code
        regsub -all {([^%])%X} $source_code "\\1$xml"                source_code
        regsub -all {([^%])%N} $source_code "\\1$node"               source_code
        regsub -all {([^%])%I} $source_code "\\1$a_args(-widget)"    source_code
        regsub -all {%%}       $source_code %                        source_code

        #catch {eval $source_code} res
        set res [eval $source_code]
        if { [string index $res 0] == "." } {
          if { [llength $res] == 2 } {
            return [list $wParent.$a_args(-widget) [lindex $res 1] ]
          } else {
            return [list $wParent.$a_args(-widget) {}]
          }
        } else {
          return [list $wParent.$a_args(-widget) $res]
        }
      } ;# CREATECODE
    } ;# switch

  } ;# proc createWidget


  proc data { action var args } {
    variable DATA
    # Standard Data:
    #  * currentTextWidget
    #  * currentTextXML
    #  * guiXML
    #  * tkXMLiVEversion

    switch -- $action {
      set {
        set DATA($var) [lindex $args 0]
      } ;# switch -- set
      get {
        if { $var == "currentTextXML" } {
          return [richtext::widgetData [data get currentTextWidget] get xml]
        } else {
          if { [info exists DATA($var)] } {
            return $DATA($var)
          } else {
            return -code error -errorinfo "Can't found GUI Data '$var'.\n"
          } ;# if-else (Data present)
        } ;# if currentTextXML

      } ;# switch -- get
    } ;# switch (action)
  } ;# proc data


  proc balloon { widget action args } {
  variable BALLOON

    switch -- $action {
      set {
        if { $args != {{}} } {
          balloon $widget clear
          #bind $widget <Any-Enter> "after 1000 [list gui::balloon %W show $args mousepointer %X %Y]"
          #bind $widget <Any-Leave> "catch { destroy %W.balloon }"
          bind $widget <Enter> " gui::balloon $widget show $args "
          bind $widget <Leave> " wm withdraw .bubble "
        }
      }
      show {
        if ![winfo exists .bubble] {
          toplevel .bubble -relief flat -background black -bd 1
          wm withdraw .bubble
          update
          array set attrFont [font actual fixed]
          set attrFont(-size) [expr $attrFont(-size) - 2]
          eval pack [message .bubble.txt -aspect 5000 -bg lightyellow \
                                 -font [array get attrFont] -text [lindex $args 0]]
          pack .bubble.txt
          wm transient .bubble .
          wm overrideredirect .bubble 1
          bind .bubble <Enter> "wm withdraw .bubble"
        } ;# if

        if {$args == ""} {  wm withdraw .bubble  }
        set text [lindex $args 0]

        set BALLOON $text
        switch $text {
          ""        {   wm withdraw .bubble ; update  }
          "default" {
            after 1000 "gui::raise_balloon $widget {$text}"
            after 7000 "if { \$gui::BALLOON == {$text} } { wm withdraw .bubble ; update }"
          }
        } ;# switch
      }
      clear {
        catch { destroy .balloon }
        bind $widget <Enter> {}
        bind $widget <Leave> {}
      }
    } ;# switch action
  } ;# proc balloon


  proc raise_balloon {widget text} {
  variable BALLOON

    if { $BALLOON != $text } {   wm withdraw .bubble ; update ; return  }
    set cur_widget [winfo containing [winfo pointerx .] [winfo pointery .]]
    if { $cur_widget != $widget } {  return  }

    raise .bubble
    .bubble.txt configure -text $text
    set b_x [expr [winfo pointerx .] - [winfo reqwidth .bubble]/2]
    set b_y [expr [winfo pointery .] + 15]
    wm geometry .bubble +$b_x+$b_y
    wm deiconify .bubble
    update
  } ;# proc raise_balloon


  proc clipboardGet { args } {
    set a_args(-selection)  CLIPBOARD
    set a_args(-encoding)   [encoding system]
    array set a_args $args

    if { [catch {::selection get} sel] } {
      if { [catch {::selection get -selection $a_args(-selection)} sel ] } {
        return ""
      } ;# if
    } ;# if

    if { [string equal [::selection own] {}] } {
      set sel [encoding convertfrom $a_args(-encoding) $sel]
    } ;# is selection own

    return $sel
  } ;# proc clipboardGet


  proc Dialog { xml varName } {
    variable dialogVar
    set dialogVar($varName) {}

    if { [string equal $xml {}] } { set xml [data get guiXML] } ;# if $xml == ""
    set f .dialog_$varName
    if { [winfo exists $f] }  {  destroy $f  } ;# if Dialog exists
    toplevel $f

    set d_node {}
    foreach { node } [simxml::getNodes $xml {} name DIALOG] {
      catch {
        if { [simxml::attribute $xml get $node -name id] == $varName } {
          set d_node $node  ;   break
        } ;# if
      } ;# catch
    } ;# foreach

    array set rAttr [simxml::attribute $xml get $d_node -param true]
    set rWidget  $rAttr(-id)    ;  unset rAttr(-id)
    catch { set tWidget  $rAttr(-title) ;  unset rAttr(-title) }

    foreach { Widget } [simxml::nodeInfo $xml $d_node children] {
      set paneWidget [createWidget $f $Widget -widget dlg]
    } ;# foreach

    catch { wm title $f [mc $tWidget] }
    wm withdraw $f
    update idletasks
    set x [expr [winfo screenwidth  $f]/2 - [winfo reqwidth  $f]/2 - [winfo vrootx [winfo parent $f]]]
    set y [expr [winfo screenheight $f]/2 - [winfo reqheight $f]/2 - [winfo vrooty [winfo parent $f]]]
    wm geom $f +$x+$y
    wm deiconify $f

    grab $f
    tkwait window $f

    return $dialogVar($varName)
  } ;# proc Dialog

  proc getStyle { xml style } {
  variable xmlNODE
    array set rAttr {}

    set t [time {set styles_node [simxml::getNodes $xml {} name STYLES]} 1]
    #set t [time {set styles_node $xmlNODE(STYLES)} 1]

    foreach { st_child } [simxml::getNodes $xml $styles_node name ITEM] {
      if {[string equal [simxml::attribute $xml get $st_child -name id] $style] } {
        array set rAttr [simxml::attribute $xml get $st_child -param true]
        catch { unset rAttr(-style) }
        catch { unset rAttr(-id)    }
        break
      } ;# if
    } ;# foreach

  return [array get rAttr]
  } ;# proc getStyle


  proc menuCreate { widget xml id } {
  variable menuINDEX

    set style [ set menuNode {} ]
    set root_node [simxml::getNodes $xml {} name MENUS -one-level-child true]
    foreach { menuNodes } [simxml::nodeInfo $xml $root_node children] {
      if {[string equal [simxml::attribute $xml get $menuNodes -name id] $id] } {
        set menuNode $menuNodes ; break
      } ;# if id == id
    } ;# foreach all MENU nodes
    if { [string equal $menuNode {}] } { return -code error }

    catch { set style [gui::getStyle $xml \
                      [simxml::attribute $xml get $menuNodes -name style]]
    } ;# catch
    if { $widget == "." } { set w "" } else { set w $widget }

    set rootMenu [eval menu $w.[simxml::attribute $xml get $menuNode -name id] $style]
    $widget config -menu $rootMenu


    foreach { menuNodes } [simxml::nodeInfo $xml $menuNode children] {
      menuAdd $rootMenu $xml $menuNodes $style
    } ;# foreach element in founded menu node
  } ;# proc menuCreate


  proc menuAdd { widget xml menuNode args } {
  variable menuINDEX
    array set attr [simxml::attribute $xml get $menuNode]

    foreach { item } [list name      label  accel    command  state \
                           underline value  variable onvalue  offvalue] {
      set v$item {}  ;  catch { set v$item $attr($item) }
    } ;# foreach initialize variable

    set vlabel [mc $vlabel]

    catch { set vcommand_other [simxml::getText $xml $menuNode] }
    if { ! [string equal $vcommand_other {}] } { set vcommand $vcommand_other }

    switch -- [simxml::nodeInfo $xml $menuNode name] {
      CASCADE    -
      MENUBUTTON {
        set wMenuButton [eval menu $widget.$vname [lindex $args 0]]
        $widget add cascade -label $vlabel -menu $wMenuButton
        catch { $widget entryconfigure $vlabel -underline $vunderline }
        set menuINDEX($widget||$vname) [$widget index $vlabel]

        foreach { menuNodes } [simxml::nodeInfo $xml $menuNode children] {
          menuAdd $wMenuButton $xml $menuNodes [lindex $args 0]
        } ;# foreach element in founded menu node
      } ;# MENUBUTTON
      RADIO    -
      CHECK    -
      COMMAND  {
        catch { bind . <$vaccel> "$vcommand ; break" }

        array set listKeys [list DELETE DEL INSERT INS DOLLAR \$]
        if { [regexp {^(.+)-([^+]+)$} $vaccel -> vControl vKey] } {
          switch -- $vControl {
            Control { set vaccel "Ctrl"  }
            Shift   { set vaccel "Shift" }
          }
          set vKey [string toupper $vKey]
          catch { set vKey $listKeys($vKey) }
          set vaccel "$vaccel+$vKey"
        } ;# if

        set wType [string tolower [simxml::nodeInfo $xml $menuNode name] ]
        $widget add $wType -label $vlabel -command $vcommand -accelerator $vaccel
        set menuINDEX($widget||$vname) [$widget index $vlabel]
        catch { $widget entryconfigure $vlabel -state     $vstate     }
        catch { $widget entryconfigure $vlabel -underline $vunderline }
        catch { $widget entryconfigure $vlabel -value     $vvalue     }
        catch { $widget entryconfigure $vlabel -variable  $vvariable  }
        catch { $widget entryconfigure $vlabel -onvalue   $vonvalue   }
        catch { $widget entryconfigure $vlabel -offvalue  $voffvalue  }
      } ;# ITEM
      SEPARATOR {
        $widget add separator
      } ;# SEPARATOR
    } ;# switch
  } ;# proc menuAdd


  proc menuGet { widget args } {
  variable menuINDEX

  set result {}

  if { [llength $args] == 1 } {
    set result $menuINDEX($widget||[lindex $args 0])
  } else {
    switch -- [lindex $args 0] {
      -label { set result [ $widget index [lindex $args 0] ] }
      -node {      }
      } ;# switch
  } ;# if -else

  return $result
  } ;# proc menuGet


  proc menuDel { widget args } {
  variable menuINDEX

  if { [llength $args] == 1 } {
    $widget delete [menuGet $widget [lindex $args 0]]
  } else {
    switch -- [lindex $args 0] {
      -label { $widget delete [lindex $args 0] }
      -node  {      }
      } ;# switch
  } ;# if-else
  } ;# proc menuDel

  proc toolbarCreate { parent args } {
    variable ID
    variable toolBar
    set a_args(-file) {}
    set a_args(-xml)  {}
    array set a_args $args

    if { [string equal $a_args(-xml) {}] } {
      set xml  [simxml::parseXML -file $a_args(-file)]
    } else {
      set xml $a_args(-xml)
    }

    foreach { tbNode } [simxml::getNodes $xml {} name TOOLBAR ] {
      set tbWidget tb:[simxml::attribute $xml get $tbNode -name id]
      set toolBar($tbWidget) [frame $parent.$tbWidget -height 32 -borderwidth 0 -relief groove -background #EEEEEE]

      foreach { menus } [simxml::nodeInfo $xml $tbNode children] {
        switch -- [simxml::nodeInfo $xml $menus name] {
          SEPARATE {
            incr ID
            canvas $toolBar($tbWidget).separate$ID -height 32 -width 3 -relief flat -borderwidth 0 -highlightbackground #eeeeee
            pack $toolBar($tbWidget).separate$ID -side left
          }
          BUTTON {
            set bCommand [ set bText [ set bImage {} ] ]
            set bWidget [simxml::attribute $xml get $menus -name id]
            if { ! [string equal [set nodeImg [simxml::getNodes $xml $menus name IMAGE]] {}] } {
              set bImage [simxml::attribute $xml get $nodeImg -name src]
              set bImage [list -image [image create photo -file $bImage]]
            } ;# if
            if { ! [string equal [set nodeTxt [simxml::getNodes $xml $menus name TEXT]] {}] } {
              set bText [list -text [mc [simxml::getText $xml $nodeTxt]] ]
            } ;# if
            if { ! [string equal [set nodeCmd [simxml::getNodes $xml $menus name COMMAND]] {}] } {
              set bCommand [simxml::getText $xml $nodeCmd]
            }
            eval {button $toolBar($tbWidget).$bWidget -command $bCommand} $bImage $bText

            gui::balloon $toolBar($tbWidget).$bWidget set \
                      [mc [simxml::getText $xml [simxml::getNodes $xml $menus name POPUP]] ]

            pack $toolBar($tbWidget).$bWidget -side left
          }
        } ;# switch
      } ;# foreach ALL BUTTON, LIST, SEPARATE ans others
      pack $toolBar($tbWidget) -side top -fill x
    } ;# foreach TOOLBAR node
  } ;# proc toolbarCreate


  proc msgDebug { text } {
    catch { [gui::data get debugWidget].txt insert end "$text\n" 
            [gui::data get debugWidget].txt see end
          }
    puts stderr $text
  } ;# proc msgDebug

} ;# namespace gui




