package provide simxml 1.0

namespace eval simxml {
  variable xmlDOM        ;  array set xmlDOM       {}
  variable XML 0
  variable ID  0
  variable xmlENTITY     ;  array set xmlENTITY    {}
  variable xmlPrCommand  ;  array set xmlPrCommand {}
  variable xmlDTD        ;  array set xmlDTD       {}
  variable fastNODE      ;  array set fastNODE     {}


  proc delete { xml } {
    variable xmlDOM

    foreach { key } [array names xmlDOM "$xml:*"] {  catch { unset xmlDOM($key) }  } ;# foreach
    catch { unset xmlDOM($xml) }
    foreach { key } [array names xmlENTITY "$xml:*"] {  catch { unset xmlENTITY($key) }  } ;# foreach
    catch { unset xmlENTITY($xml) }
    foreach { key } [array names xmlDTD "$xml:*"] {  catch { unset xmlDTD($key) }  } ;# foreach
    catch { unset xmlDTD($xml) }
    foreach { key } [array names xmlPrCommand "$xml:*"] {  catch { unset xmlPrCommand($key) }  } ;# foreach
    catch { unset xmlPrCommand($xml) }

  } ;# proc delete

  proc create { args } {
    variable xmlDOM
    variable xmlDTD
    variable XML

    set a_args(-widget)  ""
    set a_args(-extra)   ""
    set a_args(-text)    ""
    set a_args(-version)  "1.0"
    set a_args(-encoding) [encoding system]
    set a_args(-file)    ""
    array set a_args $args

    incr XML

    set xml $XML
    if { [info exists xmlDOM($xml)] } {
      foreach index [array names xmlDOM "$xml*"] {
        catch { unset xmlDOM($index) }
      }
    } ;# if

    set xmlDOM($XML)               {}
    set xmlDOM(${XML}_encoding)    $a_args(-encoding)
    set xmlDOM(${XML}_version)     $a_args(-version)
    set xmlDOM(${XML}_file)        $a_args(-file)
    set xmlDTD($XML)               none

    return $XML
  } ;# proc create

  proc xmlInfo { xml action args } {
    variable xmlDOM

    switch -- $action {
      get {
        foreach { name } $args {
          set result $xmlDOM(${xml}_$name)
        } ;# foreach
        return $result
      }
    } ;# switch
  } ;# proc xmlInfo

  proc add { xml parent node args } {
    variable fastNODE
    variable xmlDOM
    variable ID

    if { ![info exists xmlDOM($xml)] } {
      return -code error -errorinfo "XML DOM is not created.\n"
    } else {
      if { [string equal $parent {}] } {
        set parent [root get $xml]
      } ;# if (parent == "")
    } ;# if-else

    set a_args(-type)      "TAG"
    set a_args(-widget)    {}
    set a_args(-position)  "end"
    set a_args(-attribute) {}
    set a_args(-text)      {}
    set a_args(-extra)     {}
    array set a_args $args

    set node [string toupper $node]

# if (New TAG is not a first in DOM structure)
if { ![string equal $parent {}] } {
    foreach { pos_command pos_value } $a_args(-position) { break; }
    set a_args(-double-text) 0
    incr ID

    if { $pos_value != {} } {
      set parent   $pos_value
    } ;# if


# SELECT position for TAG
    switch -- $pos_command {
      before {
        set parent   [nodeInfo $xml $pos_value parent]
        set in_index [lsearch $xmlDOM($xml:$parent:children) $pos_value]
        set pos_text 0
      } ;# before
      after {
        set parent   [nodeInfo $xml $pos_value parent]
        set in_index [expr [lsearch $xmlDOM($xml:$parent:children) $pos_value] + 1]
        set pos_text -1
      } ;# after
      start {
        set in_index 0
        set pos_text 0
      } ;# start
      end {
        set in_index [llength $xmlDOM($xml:$parent:children)]
        set pos_text -1
      }
    } ;# switch -position


    if { ![eval {canAdd $xml $parent $node} [array get a_args]] } {
      # BUG NEED eval Sys event with dialog
      #return -code error -errorcode 10 -errorinfo "Can't add new tag '$node' for parent '$parent' in '$xml'.\n"
      return -1
    }

# Concatenate TEXT tag if needed
   if { [string equal $a_args(-type) TEXT] } {
     foreach { direction } {-1 0} {
       set node_tag [lindex $xmlDOM($xml:$parent:children) [expr $in_index  + $direction]]
       if { $node_tag != {} } {
         if { [string equal [nodeInfo $xml $node_tag type] "TEXT"] } {
           set text_tag [nodeInfo $xml $node_tag text]
           set concat_text "$text_tag$a_args(-text)"
           if { $direction == 0 } {
             set concat_text "$a_args(-text)$text_tag"
           } ;# if

           setParam $xml $node_tag -text $concat_text
           return $node_tag
        } ;# if (node_tag == TEXT)
       } ;# if (node_tag != "")
     } ;# foreach direction
   } ;# if TAG is TEXT

   set xmlDOM($xml:$parent:children) [linsert $xmlDOM($xml:$parent:children) \
                                              $in_index \
                                              $ID]
}

# SET TAG variable
    set xmlDOM($xml:$ID:name)      $node
    set xmlDOM($xml:$ID:type)      $a_args(-type)
    set xmlDOM($xml:$ID:widget)    $a_args(-widget)
    set xmlDOM($xml:$ID:text)      $a_args(-text)
    set xmlDOM($xml:$ID:attribute) $a_args(-attribute)
    set xmlDOM($xml:$ID:parent)    $parent
    set xmlDOM($xml:$ID:children)  {}
    set xmlDOM($xml:$ID:extra)     $a_args(-extra)        ;# additional attribute

# if TAG in first in DOM structure set ROOT element
    if { ( [string equal [root get $xml] {}]) &&
         (![string equal $a_args(-type) COMMENT]) } {
      root set $xml $ID
    }

    set fastNODE($node) [lappend fastNODE($node) $ID]
    return $ID
  } ;# proc add


  proc isNodeExists { xml node } {
    variable xmlDOM
    return [info exists xmlDOM($xml:$node:name)]
  } ;# proc isNodeExists


  proc setParam { xml node args } {
    variable xmlDOM
    array set a_args $args

    if { [isNodeExists $xml $node] } {
      foreach { param value } [array get a_args] {
        set param [string range $param 1 end]
        if {[string equal $param "text"]} {
          if {! [string equal [nodeInfo $xml $node extra] CDATA] } {
            regsub -all {\s+} $value " " value
          } ;# if CDATA
        }
        set xmlDOM($xml:$node:$param) $value
      } ;# foreach
    } else {
      return -code error -errorcode 11 -errorinfo "Tag $node in $xml not found.\n"
    } ;# if-else isNodeExists
  } ;# proc setParam

  proc children { xml node args } {
    variable xmlDOM
# VERBOSE may be FALSE | true | expanded
    set a_args(-verbose) "false"
    set a_args(-position) ""
    array set a_args $args

    set tmp {}
    set lChildren $xmlDOM($xml:$node:children)
    if { [string equal $a_args(-verbose) true] } {
      foreach { children } $lChildren  {
        lappend tmp [nodeInfo $xml $children name]
      } ;# foreach
    set lChildren $tmp ; return $lChildren
    } ;# if
    if { [string equal $a_args(-verbose) expanded] } {
      foreach { children } $lChildren  {
        lappend tmp $children [nodeInfo $xml $children name]
      } ;# foreach
    set lChildren $tmp ; return $lChildren
    } ;# if

    if { ! [string equal $a_args(-position) {}] } {
      return [lsearch -exact [nodeInfo $xml $node children] \
                             $a_args(-position)]
    } ;# if (POSITION)
  } ;# proc children


  proc nodeInfo { xml node type args } {
    variable xmlDOM
    set a_args(-no-error)  false
    array set a_args $args

    set data {}
    if { ![info exists xmlDOM($xml:$node:$type)] } {
      if { $a_args(-no-error) == "true" } {
        return {}
      } else {
        return -code error -errorinfo "Parametr '$type' not found in tag '$node' for '$xml'.\n";
      } ;# if -no-error
    }

    return $xmlDOM($xml:$node:$type)
  } ;# proc nodeInfo


  proc attribute { xml action node args } {
    variable xmlDOM

    set vAttribute       {}
    set a_args(-name)    {}
    set a_args(-value)   {}
    set a_args(-verbose) {}
    set a_args(-param)   {}
    array set a_args   $args

    switch -- $action {
# Get specific or all attributes from tag
      get {
        if { ![string equal $a_args(-name) {}] } {
          foreach { param value } [nodeInfo $xml $node attribute] {
            if { [string equal $a_args(-name) $param] } {
              return $value
            }
          } ;# foreach
          return -code error -errorinfo "Can't found attibute ($a_args(-name)) in '$xml'.\n"
        } else {
          ;# Verbose true mode
          if { [string equal $a_args(-verbose) true] } {
            foreach { param value } [nodeInfo $xml $node attribute] {
              append vAttribute " $param=\"$value\""
            } ;# foreach
            return $vAttribute
          } elseif { [string equal $a_args(-param) true] } {
            foreach { param value } [nodeInfo $xml $node attribute] {
              lappend vAttribute -$param $value
            } ;# foreach
            return $vAttribute
          } else {
            return [nodeInfo $xml $node attribute]
          } ;# if
        } ;# if
      } ;# get
# Count number of attributes for tag
      count {
        return [llength [nodeInfo $xml $node attribute]]
      } ;# count
# Set attribute for tag
      set {
        if { ![string equal $a_args(-name) {}] } {
          lappend xmlDOM($xml:$node:attribute) $a_args(-name) $a_args(-value)
        } else {
          return -code error -errorinfo \
                 "Not specific attribute name for tag \
                 '[nodeInfo $xml $node name]'.\n"
        } ;# if
      } ;# set
      del {
# Delete attribute from tag
        set tmp_attr {}
        set is_found 0

        if { ![string equal $a_args(-name) {}] } {
          foreach { param value } [nodeInfo $xml $node attribute] {
            if { ![string equal $a_args(-name) $param] } {
              lappend tmp_attr $param $value
            } else {
              incr is_found
            } ;# if
          } ;# foreach
          if { $is_found == 0 } {
            return -code error -errorinfo \
                 "Can't delete. Not found attribute '$a_args(-name)' \
                 for tag '[nodeInfo $xml $node name]'.\n"
          } ;# if

          set xmlDOM($xml:$node:attribute) $tmp_attr
        } else {
          return -code error -errorinfo \
                 "Can't delete. Not specific attribute name for tag \
                 '[nodeInfo $xml $node name]'.\n"
        } ;# if
      } ;# del
    } ;# switch

  } ;# proc attribute


  proc debug { xml node args } {
    variable xmlDOM

    set a_args(-parent)  [root get $xml]
    set a_args(-level)   0
    array set a_args $args

    set text {}
    set space [string repeat "  " $a_args(-level)]

    if { [string equal $node {}] } {
      set node $a_args(-parent)
    } ;# if ($node eq "")


    if { $a_args(-level) == 0 } {
      append text "<?xml version=\"$xmlDOM(${xml}_version)\" encoding=\"$xmlDOM(${xml}_encoding)\"?>\n\n"
      catch {
        foreach { tmp } [simxml::PrCommand $xml get XML-STYLESHEET] {
          array set tmp_style $tmp
          append text "<?xml-stylesheet type=\"$tmp_style(type)\" href=\"$tmp_style(href)\"?>\n"
        } ;# foreach XML-STYLESHEET
      } ;# catch
    } ;# if level == 0


# OPEN Tag
    switch -- [nodeInfo $xml $node type] {
      TAG {
        append text $space
        append text "<[nodeInfo $xml $node name][attribute $xml get $node -verbose true]>"
      } ;# TAG
      ELEMENT {
        append text $space
        append text "<[nodeInfo $xml $node name][attribute $xml get $node -verbose true]/>"
#        append text "\t# ID=$node |[nodeInfo $xml $node widget]|"
        append text "\n"
      } ;# ELEMENT
      TEXT {
        set tmp_text [nodeInfo $xml $node text]
        if { [lsearch [nodeInfo $xml $node extra] CDATA] != -1 } {
          append text {<![CDATA[} $tmp_text {]]>}
        } else {
          append text $tmp_text
        } ;# if-else EXTRA CDATA parameter
      } ;# TEXT
      COMMENT {
        append text "<!-- [nodeInfo $xml $node text] -->\n"
      }
    } ;# switch

# RECURSE
    foreach { tmp_node } [nodeInfo $xml $node children] {
       append text "[debug $xml $tmp_node -level [expr $a_args(-level) + 1]]"
    } ;#foreach

#CLOSE Tag
    if { [string equal [nodeInfo $xml $node type] TAG] } {
#        append text $space
        append text "</[nodeInfo $xml $node name]>"
#        append text "\t#ID=$node |[nodeInfo $xml $node widget]|"
        append text "\n"
    } ;# if (node eq "TAG")

    return $text
  } ;# proc debug


  proc root { action xml args } {
    variable xmlDOM

    switch -- $action {
      get {
        if [info exists xmlDOM($xml)] {
          return $xmlDOM($xml)
        } else {
          #BUG need error code
          return {}
        }
      } ;# get
      set {
        set xmlDOM($xml) [lindex $args 0]
      } ;# set
    } ;# switch
  } ;# proc root


# Look in DTD for check trust ADD Tag into other TAG group
  proc canAdd { xml parent node args } {
    variable xmlDOM
    variable xmlDTD
    set a_args(-1)   on
    catch { array set a_args $args }

    set parent [nodeInfo $xml $parent name]
    if { [string equal $node {}] } { set node {#PCDATA} }

    if [string equal $xmlDTD($xml) none] {
      if { [string equal $a_args(-type) ELEMENT] } {
        if { ! [info exists xmlDTD($xml:$node)] } {
          set xmlDTD($xml:$node) EMPTY
        }
        if [info exists xmlDTD($xml:$parent:_list)]  {
          if { [lsearch  $xmlDTD($xml:$parent:_list) $node] == -1 } {
            lappend xmlDTD($xml:$parent:_list) $node
            set xmlDTD($xml:$parent:_list) [lsort $xmlDTD($xml:$parent:_list)]
          } ;# if exists
        } else {
          set xmlDTD($xml:$parent:_list) $node
        } ;# if-else
        set xmlDTD($xml:$parent:_list) [lsort $xmlDTD($xml:$parent:_list)]
      } else {
        set xmlDTD($xml:$parent) ALL
        if [info exists xmlDTD($xml:$parent:_list)]  {
          if { [lsearch  $xmlDTD($xml:$parent:_list) $node] == -1 } {
            lappend xmlDTD($xml:$parent:_list) $node
            set xmlDTD($xml:$parent:_list) [lsort $xmlDTD($xml:$parent:_list)]
         } ;# if SEARCH node
        } else {
          set xmlDTD($xml:$parent:_list) $node
        } ;# if-else exists XML:_list
      } ;# if-else ELEMENT
      return 1
    } ;# if NONE

    return 0
  } ;# proc canAdd

  proc showDTD { xml } {
    variable xmlDTD

  set res {}
    foreach { key } [array names xmlDTD "$xml:*"] {
      regexp {.*:(.*?)$} $key -> element
      switch -- $xmlDTD($key) {
        EMPTY {  append res "<!ELEMENT $element EMPTY>\n" }
        ALL   {  append res "<!ELEMENT $element ([join $xmlDTD(${key}:_list) " | "])*>\n" }
      } ;# switch
    } ;# foreach
    return $res
  } ;# proc showDTD


  proc dtdListTag { xml parent args } {
    variable xmlDTD

    set res {}
    foreach { key } [array names xmlDTD "$xml:$parent"] {
      switch -- $xmlDTD($key) {
        EMPTY {  return {} }
        ALL   {  eval lappend res _ALL_ $xmlDTD(${key}:_list) }
      } ;# switch
    } ;# foreach
    return $res
  } ;# proc dtdListTag

  proc dtdTypeTag { xml parent args } {
    variable xmlDTD

    foreach { key } [array names xmlDTD "$xml:$parent"] {
      switch -- $xmlDTD($key) {
        EMPTY   {  return ELEMENT }
        ALL     {  return TAG     }
        #PCDATA {  return TEXT    }
      } ;# switch
    } ;# foreach
    # BUG!!!!!!!!!!!!! TEXT type
    return TAG
  } ;# proc dtdTypeTag

  proc listNodePath { xml node args } {
    variable xmlDOM

    set a_args(-verbose)     "false"
    set a_args(-only-parent) "false"
    set a_args(-for-css)     "false"
    set a_args(-root)        {}
    set a_args(-expanded)    "false"
    array set a_args $args

    set root 0

    if { $a_args(-for-css) == "true" } {
      return [listNodePath $xml $node -verbose true]
    } ;# if FOR-CSS is true
    if { $a_args(-only-parent) == "false" } {
      lappend lPath $node
    } else {
      set lPath {}
      set vPath {}
    }

    if { [string equal $node {}] } { return {} }

    while { 1 } {
      set parent [nodeInfo $xml $node parent]
      if { [string equal $parent ""] }   { break }
      set lPath [linsert $lPath 0 $parent]
      if { $parent == $a_args(-root) } { break }
      set node $parent
    } ;# while (infinity)

# Check Verbose mode
    if { [string equal $a_args(-verbose) true] } {
      foreach { node } $lPath {
        if { $a_args(-expanded) == "false"} {
          lappend vPath [nodeInfo $xml $node name]
        } else {
          lappend vPath [list [nodeInfo $xml $node name] $node]
        } ;# if expanded
      } ;# foreach
      set lPath $vPath
    } ;# if (verbose wq TRUE)

    return $lPath
  } ;# proc listNodePath


# Get ID node list for TAG with needed param value
  proc getNodes { xml root param value args} {
    variable fastNODE
    variable xmlDOM
    set a_args(-no-child)        false
    set a_args(-one-level-child) false
    array set a_args $args

    if { ([string equal $root {}]) && ($a_args(-no-child) == "false") && $param == "name"} {
      return $fastNODE($value)
    } ;# if

    if { [string equal $root {}] } { set root [root get $xml] } ;# if

    set result {}
    set expres "$xml:$root:$param"
    foreach { key val } [array get xmlDOM $xml:$root:$param] {
      if { [string equal $val $value] } {  lappend result $root  } ;# if
    } ;# foreach

    if { [string equal $a_args(-one-level-child) true] } \
       {  set no_child true  } else {  set no_child false  } ;# if

    if { [string equal $a_args(-no-child) false] } {
      foreach { child } [nodeInfo $xml $root children] {
        set tmp_result [getNodes $xml $child $param $value -no-child $no_child]
        if { ! [string equal $tmp_result {}] } { eval lappend result $tmp_result } ;# if
      } ;# foreach children for $root
    } ;# if -no-child

    return $result
  } ;# proc getNodes


  proc nodeTextConcatenate { xml node } {
    variable xmlDOM

    set last_node      {}
    set last_node_type {}
    foreach { child } [nodeInfo $xml $node children] {
      set child_type [nodeInfo $xml $child type]
      if { [string equal $child_type TEXT] } {
        if { [string equal $last_node_type TEXT] } {
          set text_child [nodeInfo $xml $child text]
          setParam $xml $last_node -text "[simxml::nodeInfo $xml $last_node text]$text_child"
          del $xml $child
          nodeTextConcatenate $xml $node
          break
        } ;# if last_node is TEXT
      } ;# if (child is TEXT)
    set last_node      $child
    set last_node_type $child_type
    } ;# foreach ALL CHILD
  } ;# proc nodeTextConcatenate


  proc getText { xml node args } {
    set result {}
    foreach { child } [simxml::getNodes $xml $node type TEXT -one-level-child true] {
      lappend result [simxml::nodeInfo $xml $child text]
    } ;# foreach

    return [join $result $args]
  } ;# proc getText


  proc getALL { xml } {
    variable xmlDOM
    variable xmlENTITY

    set text {}
    foreach { key } [lsort [array names xmlDOM]] {
      append text "$key = \t$xmlDOM($key)\n"
    } ;# foreach

    append text "===================================\n"
    foreach { key } [lsort [array names xmlENTITY]] {
      append text "$key = \t$xmlENTITY($key)\n"
    } ;# foreach
    return $text
  } ;# proc getALL

  proc del { xml node_root args } {
    variable xmlDOM

    set children [nodeInfo $xml $node_root children]

    if { ! [string equal $children {}] } {
    foreach { child } $children {
      del $xml $child
    } ;# foreach (CHILDREN)
    } ;# if (TAG have'nt children)

    set parent   [nodeInfo $xml $node_root parent]
    set position [children $xml $parent -position $node_root]
    setParam $xml $parent -children \
                                               [lreplace \
                                                 [nodeInfo \
                                                                       $xml \
                                                                       $parent \
                                                                       children] \
                                                 $position \
                                                 $position \
                                               ]

    # BUG need delete All attribute for TAG
    foreach { param } [list name type attribute parent children widget extra text] {
      unset xmlDOM($xml:$node_root:$param)
    } ;# foreach
    nodeTextConcatenate $xml $parent
  } ;# proc del


  proc parseXML { args } {
    variable xmlDOM
    variable xmlPrCommand

    set a_args(-file)   {}
    set a_args(-text)   {}
    array set a_args $args

    if { ! [string equal $a_args(-file) {}] } {
      set FILE [open $a_args(-file) RDONLY]
        fconfigure $FILE -encoding binary
        set source_xml [read $FILE]
      close $FILE
    } elseif {! [string equal $a_args(-text) {}] } {
      set source_xml $a_args(-text)
    } else {
      return -code error -errorcode 10 -errorinfo "parseXML bad parameters\n"
    } ;# if-else


# Now we can Parse XML document

    # Read XML version and other attribute
    set attribute [ set xml_attribute {} ]
    array set aAttr [list -encoding [encoding system]]
    regexp {^(.*?)\n} $source_xml -> fLine
    regexp {^<\?xml\s*(.*?)\s*?\?>} $fLine found xml_attribute
    set attribute [parseAttribute $xml_attribute -param true]
    array set aAttr $attribute
    set aAttr(-encoding) [string tolower $aAttr(-encoding)]
    switch -- $aAttr(-encoding) {
      windows1251  -
      windows-1251 { set aAttr(-encoding) cp1251 }
      utf8         { set aAttr(-encoding) utf-8  }
      utf-16       { set aAttr(-encoding) utf-16 }
    } ;# switch
# BUG!!! need proverit' encoding v XML documente na validnost'.
    set source_xml [encoding convertfrom $aAttr(-encoding) $source_xml]

# Set default value for process variable
    set pos 0                  ;# Position search Cursor
    set buffer {}              ;# Buffer for text
    set action {findXMLstart}  ;# Current operation
    set find_start 0           ;# Index for start action
    set find_end   0           ;# Index for end action
    set total_length [string length $source_xml]  ;# Total length XML document
    set nodes_list {}          ;# List of inserted nodes
    set count_nodes 0
    set count_text  0
# Type action:
#   getText
#   findXMLstart
#   findXMLend
#   findCDATAend

set xml [eval {create -file $a_args(-file)} $attribute  ]
# Process parse ======================================
    for { } { $pos < $total_length } { incr pos } {
      set char [string index $source_xml $pos]
      switch -- $action {
        findXMLstart {
          if { $char != "<" } { continue }
          set buffer  $char
          set action  findXMLend
        } ;# switch --- findXMLstart
        findCDATAend {
          append buffer $char;
          if { [string equal [string range $buffer "end-2" end] {]]>}]} {
            set action findXMLstart
            regexp {<!\[CDATA\[(.*?)\]\]>} $buffer -> buffer
            add $xml [lindex $nodes_list end] {} -text $buffer -type TEXT -extra CDATA
            set buffer {}
          } ;# if
        }
        findXMLend {
          if { $char != ">" } {
            append buffer $char;
            if { [string equal $buffer {<![CDATA[}] } {
              set action findCDATAend
            } ;# if founded CDATA section
            continue
          } ;# if CHAR ne ">"
          set node_info [string range $buffer 1 end]
          switch -- [string index $node_info 0] {
            ? {
              set node_info [string range $node_info 1 "end-1"]
              foreach { node_name node_attr } [simxml::parseNode $node_info] { break }
              set node_name [string toupper $node_name]
              #lappend xmlPrCommand($xml:$node_name) $node_attr
              PrCommand $xml set $node_name $node_attr
            }
            ! {
              set tag_name {}
              regexp {.([^\s]+)} $node_info -> tag_name
              set tag_name [string toupper $tag_name]
              switch -- $tag_name {
                ENTITY {
                  if [regexp {\w+\s+(\w+)\s+(["'])(.*?)\2>?} $node_info -> en_name {} en_value] {
                    entity $xml set $en_name $en_value
                  } ;# if regexp true
                }
                -- {
                  if [regexp {!--(.*)--$} $node_info -> comment] {
                    simxml::add $xml [lindex $nodes_list end] {} -type COMMENT -text $comment
                  } ;# if
                }
              } ;# switch
            }
            / {
              #puts "CLOSE TAG: [string range $node_info 1 end]"
              set nodes_list [lreplace $nodes_list end end]
            }
            default {
              switch -- [string index $buffer end] {
                / {
                  set node_info [string range $node_info 0 end-1]
                  foreach { node_name node_attr } [simxml::parseNode $node_info] { break }
                  add $xml \
                                           [lindex $nodes_list end] \
                                           $node_name \
                                           -attribute $node_attr -type ELEMENT
#                  puts "ELEMENT : $node_name ($node_attr)"
                  incr count_nodes
                }
                default {
                  set node_name [ set node_attr {} ]
                  foreach { node_name node_attr } [simxml::parseNode $node_info] { break }
                  lappend nodes_list [add $xml \
                                                               [lindex $nodes_list end] \
                                                               $node_name \
                                                               -attribute $node_attr]
                  #puts "OPEN TAG: $node_name ($node_attr) \[[join $nodes_list { / }]\]"
                  incr count_nodes
                }
              } ;# switch end char in node_info
            }
          } ;# switch type of TAG (NODE, COMMENT, RULES, ENTITY, e.t.c.)

          set buffer {}
          set action getText
        } ;# switch --- findXMLend
        getText {
          if { $char != "<" } { append buffer $char ; continue }
          #set buffer [string trim $buffer "\t\n "]
          regsub -all {^\n} $buffer "" buffer
          regsub -all {\n$} $buffer "" buffer
          regsub -all {\s+} $buffer " " buffer
          if { [string trim $buffer "\t\n "] != "" } {
            add $xml [lindex $nodes_list end] {} -text $buffer -type TEXT
            #puts "TEXT: |$buffer|"
            incr count_text
          } ;# if founded text if not empty

          set buffer $char
          set action findXMLend
        } ;# switch --- getText
      } ;# switch action
    } ;# Main cycle for Parse XML document

    gui::msgDebug "COUNT NODES: $count_nodes\nCOUNT_TEXT: $count_text"
    return $xml
  } ;# proc parseXML


  proc parseNode { node_info args } {
    set a_args(-param) "false"
    array set a_args $args

    set node_name      {}
    set node_attribute {}
    regexp {^(\S+)\s*(.*)$} $node_info {} node_name node_attribute
    set attribute [parseAttribute $node_attribute -param $a_args(-param)]

    return [list $node_name $attribute]
  } ;# proc parseNode


  proc parseAttribute { attribute args } {
    set a_args(-param) "false"
    array set a_args $args

    set result {}
    set suffix {}
    if [string equal $a_args(-param) true] { set suffix "-" }

    set value [ set param [ set bracket [ set buffer {} ] ] ]
    set action findEndParam

    foreach { char } [ split $attribute ""] {
      switch -- $action {
        findEndParam {
          if { [string equal $char =] } {
            set param  [string tolower [string trim $buffer]]
            set action findStartValue
            set buffer {}
            continue
          } ;# if
          append buffer $char
        }
        findStartValue {
          if { [string match {["']} $char] } {
            set action findEndValue;
            set bracket $char
            continue
          } ;# if
        }
        findEndValue {
          if { [string equal $char $bracket] } {
            lappend result ${suffix}$param $buffer
            set buffer {}
            set action findStartParam
            continue
          } ;# if
          append buffer $char
        }
        findStartParam {
          if { ![string is space $char] } {
            set action findEndParam
            set buffer $char
            continue
          } ;# if
        }
      } ;# switch
    } ;# foreach

    return $result
  } ;# proc parseAttribute


  proc entity { xml action args } {
    variable xmlENTITY

    switch -- $action {
      get {
        if [info exists xmlENTITY($xml:[lindex $args 0])] {
          return $xmlENTITY($xml:[lindex $args 0])
        } else {
          return -code error -errorcode 10 -errorinfo "Can't found ENITITY [lindex $args 0] in '$xml'.\n"
        } ;# if
      }
      set {
        foreach { key value } $args {
          set xmlENTITY($xml:$key) $value
        } ;# foreach all ENTITY element
      }
    } ;# switch
  } ;# proc entity

  proc PrCommand { xml action args } {
    variable xmlPrCommand

    switch -- $action {
      get {
        if [info exists xmlPrCommand($xml:[lindex $args 0])] {
          return $xmlPrCommand($xml:[lindex $args 0])
        } else {
          return -code error -errorcode 10 -errorinfo "Can't found PrCommand [lindex $args 0] in '$xml'.\n"
        } ;# if
      }
      set {
        foreach { key value } $args {
          lappend xmlPrCommand($xml:$key) $value
        } ;# foreach all ENTITY element
      }
      clear {
        catch { unset xmlPrCommand($xml:[lindex $args 0]) }
      }
    } ;# switch
  } ;# proc PrCommand


  proc getTree { xml root args } {
    set a_args(-verbose)   false
    set a_args(-plain)     false
    set a_args(-command)   {}
    set a_args(-level)     0
    array set a_args $args

    if { [string equal [simxml::nodeInfo $xml $root type] TEXT] } {  return {} }
    set children [simxml::nodeInfo $xml $root children]
    if { [string equal $a_args(-verbose) true] } {
      if { [string equal $a_args(-plain) true] } {
        if { [llength $children] == 0 } {
          set child "  "
        } elseif { ([llength $children] == 1) &&
                    ([string equal [simxml::nodeInfo $xml $children type] TEXT]) } {
          set child "| "
        } else {
          set child "+ "
        }
        set result "[string repeat "  " $a_args(-level)]$child[simxml::nodeInfo $xml $root name]\n"
      } ;# if-else -plain true
    } ;# if -verbose true

    foreach { child } $children {
      append result [eval getTree $xml $child [array get a_args] -level [expr $a_args(-level) + 1]]
    } ;# foreach

    return $result
  } ;# proc getTree



  proc recurseTree { xml args } {
    variable xmlDOM

    set a_args(-root)    [root get $xml]
    set a_args(-command) {}
    set a_args(-level)   0
    array set a_args $args

    if { [string equal $a_args(-root) {}] } { set a_args(-root) [root get $xml] }
    set node $a_args(-root)

    set code $a_args(-command)
    regsub -all {([^%])%XML}      $code  "\\1$xml"       code
    regsub -all {([^%])%NODE_ID}  $code  "\\1$node"      code

    eval $code

# RECURSE
    foreach { tmp_node } [nodeInfo $xml $node children] {
       #append text "[debug $xml $tmp_node -level [expr $a_args(-level) + 1]]"
       recurseTree $xml -root $tmp_node -command $a_args(-command) \
                        -level [expr $a_args(-level) + 1]
    } ;#foreach

    return 1
  } ;# proc recurseTree

} ;# namespace simxml




